/* 
   audio_core_common.c - common stuff for, at least, OSS and ALSA cores
   
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <pthread.h>
#include <signal.h>

#include "audio_core_common.h"

void audio_core_freefrags(volatile struct audio_core_common_t * acomt)
{
	DXPRINT("enter\n");
	if(!acomt->frag)
		return;
	DXPRINT("free loop\n");
	while(acomt->nfrags > 0) { // FIXME - why does this segfault?
		acomt->nfrags--;
		DXPRINT("audio core common close - free buffer %d (%d %d).\n", acomt->nfrags, acomt->head, acomt->tail);
		if((acomt->frag + acomt->nfrags)->buf)
			free((acomt->frag + acomt->nfrags)->buf);
	}
	DXPRINT("free frag\n");
	free(acomt->frag);
	DXPRINT("leave\n");
}

ntime_t audio_core_tailtime(struct aud_core_t * audt)
{
	volatile struct audio_core_common_t * acomt = (struct audio_core_common_t *)audt->priv;

	if(!acomt)
		return 0;
	while(acomt->frag[acomt->tail].full == 0) {
		usleep(10000);
	}

	return (acomt->frag + acomt->tail)->time;
}

inline void audio_core_droptail(struct aud_core_t * audt)
{
	volatile struct audio_core_common_t * acomt = (struct audio_core_common_t *)audt->priv;
	int i;
	
	if(!acomt)
		return;
	i = acomt->tail;
	(acomt->frag + i)->full = 0;
	i ++;
	if(i >= acomt->nfrags) i = 0;
	acomt->tail = i;
}

//FIXME: will work for little-endian only... (SLD)
struct s2 {
	unsigned short lo;
	unsigned short hi;
};
typedef union {
	unsigned int ui;
	struct s2 us;
} ws_t;

int audio_core_read(struct aud_core_t * audt, ntime_t t, int drate, unsigned char * buf, int n)
{
	volatile struct audio_core_common_t * acomt = (struct audio_core_common_t *)audt->priv;
	int i, k;
	int c = 0;
	unsigned int dt = 0;
	ws_t j;
	
	/* drate is the strecthing applied */

	if(!acomt)
		return 0;
	if(drate > 0) {
		dt = drate << 15;
		dt /= 1000;  // 0 .. 32768
		dt = 65536 - dt;
	} else if(drate < 0) {
		dt = (-drate) << 15;
		dt /= 1000;  // 0 .. 32768
		dt = 65536 - dt;
		dt = (unsigned int)((unsigned long long)(1ULL<<32) / (unsigned long long)dt);
	} else {
		dt = 65536;
	}
	j.ui = 0;
	while(1) {
		i = acomt->tail;
		if((acomt->frag + i)->full == 0) {
			EPRINT("Ran out of audio data... not supposed to happen\n");
			return c;
		}
		if ((t >= 0) && (acomt->frag + i)->time > t)
			return c;
		if ((n - c) < (2 * acomt->fragsize))
			return c;
		
		if(drate == 0) {
			memcpy(buf + c, (acomt->frag + i)->buf, acomt->fragsize);
			c += acomt->fragsize;
		}else if(acomt->chans == 2) {
			unsigned int *in = (unsigned int *) ((acomt->frag + i)->buf);
			unsigned int *out = (unsigned int *) (buf + c);
			k = 0;
			// SLD: Slight tweak here: retain fractional part of j
			// between fragments.  Should help even out resampling
			// and get closer to the desired number of output
			// samples
			for(j.us.hi = 0; j.us.hi < (acomt->fragsize >> 2); j.ui += dt)
				out[k++] = in[j.us.hi];
			c += k << 2;
		} else {
			unsigned short * in = (unsigned short *) ((acomt->frag + i)->buf);
			unsigned short * out = (unsigned short *) (buf + c);
			k = 0;
			//SLD: Comment as above
			for(j.us.hi = 0; j.us.hi < (acomt->fragsize >> 1); j.ui += dt)
				out[k++] = in[j.us.hi];
			c += k << 1;
		}
		(acomt->frag + i)->full = 0;
		i ++;
		if(i >= acomt->nfrags) i = 0;
		acomt->tail = i;
	}
	return -1;
}

unsigned long long audio_core_bps(struct aud_core_t * audt)
{
	volatile struct audio_core_common_t * acomt = (struct audio_core_common_t *)audt->priv;


	if(!acomt)
		return 0;

	return acomt->bps;
}

int audio_core_has_data(struct aud_core_t * audt)
{
	volatile struct audio_core_common_t * acomt = (struct audio_core_common_t *)audt->priv;

	int diff;
        if(!acomt)
                return 0;
        diff = (acomt->head >= acomt->tail) ?
		acomt->head - acomt->tail :
		(acomt->nfrags + acomt->head) - acomt->tail;
	/* to avoid ran out of audio data */
	return diff > 10 ? acomt->fragsize * (diff - 2) : 0;
}
